import bpy
d = bpy.context.active_object.data.archipack_roof[0]
bpy.ops.archipack.material(category='roof', material='METAL')
d.tile_side = 0.0
d.hip_alt = 0.07
d.tile_fit_y = False
d.tile_space_y = 2.2
d.tile_size_z = 0.05
d.hip_size_z = 0.18
d.tile_space_x = 1.0
d.hip_size_x = 0.4
d.hip_space_x = 0.4
d.tile_enable = True
d.tile_size_x = 1.0
d.tile_border = 0.0
d.tile_bevel = False
d.tile_bevel_amt = 25.0
d.tile_solidify = False
d.tile_model = 'METAL'
d.hip_size_y = 0.18
d.tile_height = 0.02
d.tile_alternate = False
d.tile_couloir = 0.0
d.valley_enable = False
d.tile_size_y = 2.5
d.tile_altitude = 0.1
d.tile_fit_x = False
d.hip_model = 'ROUND'
d.hip_enable = False
d.tile_bevel_segs = 3
d.tile_offset = 0.0
