import bpy
op = bpy.context.active_operator

op.location = (0.0, 0.0, 0.0)
op.align = 'WORLD'
op.rotation = (0.0, 0.0, 0.0)
op.options_plus = True
op.absolute_location = False
op.use_colors = False
op.colorSet = '1'
op.random_colors = False
op.saturation = 0.75
op.geo_surface = True
op.geo_bDepth = 0.019999999552965164
op.geo_bRes = 2
op.geo_extrude = 0.0
op.geo_offset = 0.0
op.torus_p = 4
op.torus_q = 1
op.flip_p = False
op.flip_q = False
op.multiple_links = True
op.torus_u = 377
op.torus_v = 233
op.torus_rP = 0.0
op.torus_sP = 0.0
op.mode = 'MAJOR_MINOR'
op.torus_R = 1.0
op.torus_r = 0.10999999940395355
op.torus_iR = 0.75
op.torus_eR = 1.25
op.torus_s = 1.0
op.torus_h = 10.0
op.torus_res = 457
op.segment_res = 33
op.outputType = 'BEZIER'
op.handleType = 'AUTO'
op.adaptive_resolution = False
