/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The Original Code is Copyright (C) 2011 Blender Foundation.
 * All rights reserved.
 */

#include "intern/tracks.h"
#include "intern/utildefines.h"

#include "libmv/simple_pipeline/tracks.h"

using libmv::Marker;
using libmv::Tracks;

libmv_Tracks* libmv_tracksNew(void) {
  Tracks* tracks = LIBMV_OBJECT_NEW(Tracks);

  return (libmv_Tracks*)tracks;
}

void libmv_tracksDestroy(libmv_Tracks* libmv_tracks) {
  LIBMV_OBJECT_DELETE(libmv_tracks, Tracks);
}

void libmv_tracksInsert(libmv_Tracks* libmv_tracks,
                        int image,
                        int track,
                        double x,
                        double y,
                        double weight) {
  ((Tracks*)libmv_tracks)->Insert(image, track, x, y, weight);
}
