/** \file itasc/kdl/utilities/utility.cpp
 * \ingroup itasc
 */
/** @file   utility.cpp
 *  @author Erwin Aertbelien, Div. PMA, Dep. of Mech. Eng., K.U.Leuven
 *  @version 
 *      ORO_Geometry V0.2
 *   
 *  @par history
 *   - changed layout of the comments to accommodate doxygen
 */
 
#include "utility.h"

namespace KDL {

int STREAMBUFFERSIZE=10000;
int MAXLENFILENAME = 255;
const double PI=       3.1415926535897932384626433832795;
const double deg2rad = 0.01745329251994329576923690768488;
const double rad2deg = 57.2957795130823208767981548141052;
double epsilon = 0.000001;
double epsilon2 = 0.000001*0.000001;
}
