/** \mainpage Blender
 *
 * \section intro Introduction
 *
 * Blender is an integrated 3d package.
 *
 * These pages document the source code of blender.
 *
 * \subsection implinks Important Links
 * - <a href="https://developer.blender.org">developer.blender.org</a> with bug tracker.
 * - <a href="https://wiki.blender.org">Development documents</a> on our wiki.
 *
 * \subsection blother Other
 * For more information on using Blender browse to https://www.blender.org
 *
 */

/** \defgroup undoc Undocumented
 *  \brief Modules and libraries that are still undocumented, or lacking proper integration into the doxygen system, are marked in this group. */
