@title MinGW-w64 32-bit GCC build environment

@echo Setting up environment for MinGW-w64 GCC 32-bit...

@set PATH=%CD%\bin;%CD%\msys\1.0\bin;%cd%\..\..\perl32\site\bin;%cd%\..\..\perl32\bin;%cd%\..\..\c\bin;%PATH%


