
/** \defgroup render Rendering
 *  \ingroup blender
 */
/** \defgroup bmesh BMesh
 *  \ingroup blender
 */
/** \defgroup compositor Compositing */

/** \defgroup python Python
 *  \ingroup blender
 */

/** \defgroup pygen Python Generic
 *  \ingroup python
 */

/** \defgroup pymathutils Python Mathutils
 *  \ingroup python
 */

/** \defgroup pythonintern Python RNA and Operators
 *  \ingroup python
 */

/* ================================ */

/** \defgroup blender Blender */

/** \defgroup balembic BlenderAlembic
 *  \ingroup blender
 */

/** \defgroup blt BlenTranslation
 *  \ingroup blender
 */

/** \defgroup blf BlenFont
 *  \ingroup blender
 */

/** \defgroup bke BlenKernel
 *  \ingroup blender
 */

/** \defgroup bli BlenLib
 *  \ingroup blender
 */

/** \defgroup depsgraph Dependency Graph
 *  \ingroup blender
 */

/** \defgroup bph Physics
 *  \ingroup blender
 */

/** \defgroup nodes Nodes
 *  \ingroup blender
 */

/** \defgroup cmpnodes Nodes (Compositor)
 *  \ingroup nodes
 */

/** \defgroup shdnodes Nodes (Shader)
 *  \ingroup nodes
 */

/** \defgroup texnodes Nodes (Texture)
 *  \ingroup nodes
 */
/** \defgroup modifiers Object Modifiers
 *  \ingroup blender
 */

/** \defgroup shader_fx Shader Effects
 *  \ingroup blender
 */

/** \defgroup data DNA, RNA and .blend access*/

/** \defgroup gpu GPU
 *  \ingroup blender
 */

/** \defgroup ikplugin IK Plugin
 *  \ingroup blender
 */

/** \defgroup DNA Struct DNA (File Format)
 *  \ingroup blender data
 */
/** \defgroup RNA RNA (Data API)
 *  \ingroup blender data
 */

/** \defgroup blenloader Blend file IO
 *  \ingroup blender data
 *  \todo check if \ref blo and \ref blenloader groups can be
 *   merged in docs.
 */

/** \defgroup gui GUI */

/** \defgroup wm Window Manager
 *  \ingroup blender gui
 */

/* ================================ */

/** \defgroup editors Editors
 *  \ingroup blender
 */

/** \defgroup edanimation animation
 *  \ingroup editors
 */

/** \defgroup edarmature armature
 *  \ingroup editors
 */

/** \defgroup edasset asset
 *  \ingroup editors
 */

/** \defgroup edcurve curve
 *  \ingroup editors
 */

/** \defgroup eddatafiles datafiles
 *  \ingroup editors
 */

/** \defgroup edgizmolib gizmo library
 *  \ingroup editors
 */

/** \defgroup edgpencil gpencil
 *  \ingroup editors
 */

/** \defgroup edinterface interface
 *  \ingroup editors
 */

/** \defgroup edlattice lattice
 *  \ingroup editors
 */

/** \defgroup edmesh mesh
 *  \ingroup editors
 */

/** \defgroup edmeta metaball
 *  \ingroup editors
 */

/** \defgroup edobj object
 *  \ingroup editors
 */

/** \defgroup edphys physics
 *  \ingroup editors
 */

/** \defgroup edrend render
 *  \ingroup editors
 */

/** \defgroup edscr screen
 *  \ingroup editors
 */

/** \defgroup edscene scene
 *  \ingroup editors
 */

/** \defgroup edsculpt sculpt and paint
 *  \ingroup editors
 */

/** \defgroup edsnd sound
 *  \ingroup editors
 */

/** \defgroup spaction action space
 *  \ingroup editors
 */

/** \defgroup spapi space API
 *  \ingroup editors
 */

/** \defgroup spbuttons buttons space
 *  \ingroup editors
 */

/** \defgroup spconsole console space
 *  \ingroup editors
 */

/** \defgroup spfile fileselector
 *  \ingroup editors
 */

/** \defgroup spgraph graph editor
 *  \ingroup editors
 */

/** \defgroup spimage image and UV editor
 *  \ingroup editors
 */

/** \defgroup spinfo info space
 *  \ingroup editors
 */

/** \defgroup spnla NLA editor
 *  \ingroup editors
 */

/** \defgroup spnode node editor
 *  \ingroup editors
 */

/** \defgroup spoutliner outliner space
 *  \ingroup editors
 */

/** \defgroup spscript script space
 *  \ingroup editors
 */

/** \defgroup spseq sequencer
 *  \ingroup editors
 */

/** \defgroup spsnd sound space
 *  \ingroup editors
 */

/** \defgroup sptext text editor
 *  \ingroup editors
 */

/** \defgroup sptime time line
 *  \ingroup editors
 */

/** \defgroup spuserpref user preferences
 *  \ingroup editors
 */

/** \defgroup spview3d 3D view
 *  \ingroup editors
 */

/** \defgroup edtransform transform
 *  \ingroup editors
 */

/** \defgroup edutil editor utilities
 *  \ingroup editors
 */

/** \defgroup edundo undo utilities
 *  \ingroup editors
 */

/** \defgroup spuv UV editing
 *  \ingroup editors
 */

/* ================================ */

/** \defgroup editorui Interface and Widgets
 *  \ingroup gui
 */

/** \defgroup externformats External Formats */

/** \defgroup collada COLLADA
 *  \ingroup externformats
 */

/** \defgroup avi AVI
 *  \ingroup externformats
 */

/** \defgroup imbuf Image Buffer (ImBuf)
 *  \ingroup blender
 */

/** \defgroup imbcineon Cineon
 *  \ingroup imbuf
 */

/** \defgroup imbdds DDS
 *  \ingroup imbuf
 */

/** \defgroup openexr OpenEXR
 *  \ingroup imbuf
 */

/* ================================ */

/** \defgroup undoc Undocumented
 *  \brief Modules and libraries that are still undocumented, or lacking proper integration into the doxygen system, are marked in this group. */
