

// DO NOT EDIT !
// This file is generated using the MantaFlow preprocessor (prep link).

#include "vortexsheet.h"
namespace Manta {
#ifdef _C_VortexSheetMesh
static const Pb::Register _R_14("VortexSheetMesh", "VortexSheetMesh", "Mesh");
template<> const char *Namify<VortexSheetMesh>::S = "VortexSheetMesh";
static const Pb::Register _R_15("VortexSheetMesh", "VortexSheetMesh", VortexSheetMesh::_W_0);
static const Pb::Register _R_16("VortexSheetMesh", "calcCirculation", VortexSheetMesh::_W_1);
static const Pb::Register _R_17("VortexSheetMesh", "calcVorticity", VortexSheetMesh::_W_2);
static const Pb::Register _R_18("VortexSheetMesh", "reinitTexCoords", VortexSheetMesh::_W_3);
#endif
extern "C" {
void PbRegister_file_14()
{
  KEEP_UNUSED(_R_14);
  KEEP_UNUSED(_R_15);
  KEEP_UNUSED(_R_16);
  KEEP_UNUSED(_R_17);
  KEEP_UNUSED(_R_18);
}
}
}  // namespace Manta