import bpy
op = bpy.context.active_operator

op.Iterations = 1
op.IterRiver = 30
op.IterAva = 5
op.IterDiffuse = 5
op.Ef = 0.0
op.Kd = 0.10000000149011612
op.Kt = 1.0471975803375244
op.Kr = 0.009999999776482582
op.Kv = 0.0
op.userainmap = True
op.Ks = 0.5
op.Kdep = 0.10000000149011612
op.Kz = 0.30000001192092896
op.Kc = 0.8999999761581421
op.Ka = 1.0
op.Kev = 0.5
op.numexpr = True
op.Pd = 0.20000000298023224
op.Pa = 0.5
op.Pw = 1.0
op.smooth = True
op.showiterstats = False
op.showmeshstats = False
