import bpy
op = bpy.context.active_operator

op.ant_terrain_name = 'Landscape'
op.land_material = ''
op.water_material = ''
op.texture_block = ''
op.at_cursor = True
op.smooth_mesh = True
op.tri_face = False
op.sphere_mesh = False
op.subdivision_x = 128
op.subdivision_y = 128
op.mesh_size = 2.0
op.mesh_size_x = 2.0
op.mesh_size_y = 2.0
op.random_seed = 0
op.noise_offset_x = 0.0
op.noise_offset_y = 0.0
op.noise_offset_z = 0.0
op.noise_size_x = 1.0
op.noise_size_y = 1.0
op.noise_size_z = 1.0
op.noise_size = 1.25
op.noise_type = 'planet_noise'
op.basis_type = 'PERLIN_ORIGINAL'
op.vl_basis_type = 'VORONOI_CRACKLE'
op.distortion = 0.7900000214576721
op.hard_noise = '0'
op.noise_depth = 7
op.amplitude = 0.3499999940395355
op.frequency = 1.5
op.dimension = 0.9399999380111694
op.lacunarity = 2.190000057220459
op.offset = 0.5800004601478577
op.gain = 5.759998321533203
op.marble_bias = '3'
op.marble_sharp = '0'
op.marble_shape = '0'
op.height = 0.3499999940395355
op.height_invert = False
op.height_offset = 0.0
op.fx_mixfactor = 0.0
op.fx_mix_mode = '0'
op.fx_type = '0'
op.fx_bias = '0'
op.fx_turb = 0.0
op.fx_depth = 0
op.fx_amplitude = 0.5
op.fx_frequency = 1.5
op.fx_size = 1.0
op.fx_loc_x = 0.0
op.fx_loc_y = 0.0
op.fx_height = 0.5
op.fx_invert = False
op.fx_offset = 0.0
op.edge_falloff = '0'
op.falloff_x = 4.0
op.falloff_y = 4.0
op.edge_level = 0.0
op.maximum = 0.5
op.minimum = -0.5
op.vert_group = ''
op.strata = 4.0
op.strata_type = '0'
op.water_plane = False
op.water_level = 0.029999999329447746
op.remove_double = False
op.show_main_settings = True
op.show_noise_settings = True
op.show_displace_settings = True
op.refresh = True
op.auto_refresh = True
