

// DO NOT EDIT !
// This file is generated using the MantaFlow preprocessor (prep link).

#include "timing.h"
namespace Manta {
#ifdef _C_Timings
static const Pb::Register _R_16("Timings", "Timings", "PbClass");
template<> const char *Namify<Timings>::S = "Timings";
static const Pb::Register _R_17("Timings", "Timings", Timings::_W_0);
static const Pb::Register _R_18("Timings", "display", Timings::_W_1);
static const Pb::Register _R_19("Timings", "saveMean", Timings::_W_2);
#endif
extern "C" {
void PbRegister_file_16()
{
  KEEP_UNUSED(_R_16);
  KEEP_UNUSED(_R_17);
  KEEP_UNUSED(_R_18);
  KEEP_UNUSED(_R_19);
}
}
}  // namespace Manta