/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "FN_multi_function.hh"

namespace blender::fn {

class DummyMultiFunction : public MultiFunction {
 public:
  DummyMultiFunction()
  {
    static MFSignature signature = create_signature();
    this->set_signature(&signature);
  }

  static MFSignature create_signature()
  {
    MFSignatureBuilder signature{"Dummy"};
    return signature.build();
  }

  void call(IndexMask UNUSED(mask),
            MFParams UNUSED(params),
            MFContext UNUSED(context)) const override
  {
  }
};

static DummyMultiFunction dummy_multi_function_;
const MultiFunction &dummy_multi_function = dummy_multi_function_;

}  // namespace blender::fn
