import bpy
op = bpy.context.active_operator

op.bf_Model_Type = 'bf_Model_Bolt'
op.bf_Head_Type = 'bf_Head_Hex'
op.bf_Bit_Type = 'bf_Bit_None'
op.bf_Nut_Type = 'bf_Nut_Hex'
op.bf_Shank_Length = 0.0
op.bf_Shank_Dia = 10.0
op.bf_Phillips_Bit_Depth = 4.082691192626953
op.bf_Allen_Bit_Depth = 5.0
op.bf_Allen_Bit_Flat_Distance = 8.0
op.bf_Torx_Bit_Depth = 5.0
op.bf_Torx_Size_Type = 'bf_Torx_T50'
op.bf_Hex_Head_Height = 6.400000095367432
op.bf_Hex_Head_Flat_Distance = 17.0
op.bf_12_Point_Head_Height = 10.0
op.bf_12_Point_Head_Flat_Distance = 10.0
op.bf_12_Point_Head_Flange_Dia = 16.27
op.bf_CounterSink_Head_Dia = 20.0
op.bf_Cap_Head_Height = 10.0
op.bf_Cap_Head_Dia = 16.0
op.bf_Dome_Head_Dia = 20.0
op.bf_Pan_Head_Dia = 20.0
op.bf_Philips_Bit_Dia = 6.5
op.bf_Thread_Length = 20.0
op.bf_Major_Dia = 10.0
op.bf_Pitch = 1.25
op.bf_Minor_Dia = 8.646835327148438
op.bf_Crest_Percent = 10
op.bf_Root_Percent = 10
op.bf_Div_Count = 36
op.bf_Hex_Nut_Height = 8.0
op.bf_Hex_Nut_Flat_Distance = 17.0
op.bf_12_Point_Nut_Height = 10.0
op.bf_12_Point_Nut_Flat_Distance = 10.0
op.bf_12_Point_Nut_Flange_Dia = 16.27
