#define BLENDERFILE 1
#define IDR_VERSION1 1

#ifdef WINDRES
	#include "winresrc.h"
	#define IDC_STATIC (-1)
	#define STRINGIFY_(x) #x
	#define STRINGIFY(x) STRINGIFY_(x)
	#define BLEN_VER_RC_STR STRINGIFY(BLEN_VER_RC_STR_M)
#endif

APPICON ICON DISCARDABLE "winblender.ico"
BLENDERFILE ICON DISCARDABLE "winblenderfile.ico"

pointer_cursor CURSOR "cursors/pointer.cur"
moveew_cursor CURSOR "cursors/moveew.cur"
movens_cursor CURSOR "cursors/movens.cur"
arrowdown_cursor CURSOR "cursors/arrowdown.cur"
arrowleft_cursor CURSOR "cursors/arrowleft.cur"
arrowright_cursor CURSOR "cursors/arrowright.cur"
arrowup_cursor CURSOR "cursors/arrowup.cur"
cross_cursor CURSOR "cursors/cross.cur"
crossA_cursor CURSOR "cursors/crossa.cur"
crossB_cursor CURSOR "cursors/crossb.cur"
crossC_cursor CURSOR "cursors/crossc.cur"
eraser_cursor CURSOR "cursors/eraser.cur"
eyedropper_cursor CURSOR "cursors/eyedropper.cur"
handopen_cursor CURSOR "cursors/handopen.cur"
knife_cursor CURSOR "cursors/knife.cur"
pencil_cursor CURSOR "cursors/pencil.cur"
scrollew_cursor CURSOR "cursors/scrollew.cur"
scrollns_cursor CURSOR "cursors/scrollns.cur"
scrollnsew_cursor CURSOR "cursors/scrollnsew.cur"
splith_cursor CURSOR "cursors/splith.cur"
splitv_cursor CURSOR "cursors/splitv.cur"
zoomin_cursor CURSOR "cursors/zoomin.cur"
zoomout_cursor CURSOR "cursors/zoomout.cur"
forbidden_cursor CURSOR "cursors/forbidden.cur"

IDR_VERSION1 VERSIONINFO
FILEVERSION BLEN_VER_RC_1, BLEN_VER_RC_2, BLEN_VER_RC_3, BLEN_VER_RC_4
PRODUCTVERSION BLEN_VER_RC_1, BLEN_VER_RC_2, BLEN_VER_RC_3, BLEN_VER_RC_4
FILEOS 0x00000004
FILETYPE 0x00000001
BEGIN
  BLOCK "StringFileInfo"
  BEGIN
    BLOCK "04090000"
    BEGIN
      VALUE "FileVersion", BLEN_VER_RC_STR
      VALUE "ProductVersion", BLEN_VER_RC_STR
      VALUE "CompanyName", "Blender Foundation"
      VALUE "FileDescription", "Blender"
      VALUE "LegalCopyright", "GPLv3 (Blender Foundation)"
      VALUE "OriginalFilename", "blender.exe"
      VALUE "ProductName", "Blender"
    END
  END
  BLOCK "VarFileInfo"
  BEGIN
    VALUE "Translation", 0x0409, 0x0000
  END
END

